module SSG.Parser.Djot

import SSG.Parser.Core
import SSG.Parser.Util

import Data.List1
import Data.Vect
import Data.Nat
import Data.String

import Control.Eff
import Derive.Prelude

-- For iutils unit tests
import System

%language ElabReflection

--*****************************************
--*              Data Types               *
--*****************************************

public export
data Inline : Type where
  HardLineBreak : Inline
  SoftLineBreak : Inline
  NonBreakingSpace : Inline
  Text : (c : Char) -> Inline

%runElab derive "Inline" [Show, Eq]

public export
data Block : Type where

--*****************************************
--* Character Classes and String Escaping *
--*****************************************

-----------------------
-- Character classes --
-----------------------

-- Class contents

punctuationChars : List Char
punctuationChars =
  [
    '!', '"', '#', '$', '%', '&', '\''
  , '(' , ')' , '*' , '+' , ',' , '-'
  , '.' , '/' , ':' , ';' , '<' , '='
  , '>' , '?' , '@' , '[' , ']' , '^'
  , '_' , '`' , '{' , '|' , '}' , '~'
  ]

horizontalWhitespaceChars : List Char
horizontalWhitespaceChars =
 [' ', '\t']

verticalWhitespaceChars : List Char
verticalWhitespaceChars =
 ['\n', '\r']

-- Class parsers

punctuation : PS Char
punctuation = theseChars punctuationChars

--------------
-- Escaping --
--------------

escapedChar : PS Char
escapedChar = do
  _ <- thisString "\\"
  oneOfE "Expected an escapable code"
    [ punctuation
    , exactReplace "n" '\n'
    , exactReplace "t" '\t'
    , exactReplace "r" '\r'
    ]

------------------------------------
-- Line Qualifying And Whitespace --
------------------------------------

spaces : PS Nat
spaces = do
  xs <- many $ theseChars horizontalWhitespaceChars
  pure $ length xs

nonTerminal : PS Char
nonTerminal = notTheseChars verticalWhitespaceChars

lineEnding : PS String
lineEnding = theseStrings ["\r\n", "\n", "\r"]

terminal : PS ()
terminal = do
  Nothing <- tryMaybe lineEnding
    | _ => pure ()
  test <- parseEoF
  case test of
    False => throw "Expected line terminal"
    True => pure ()

line : PS (List Char)
line = do
  cs <- many nonTerminal
  terminal
  pure cs

isHorizontalWhitespace : Char -> Bool
isHorizontalWhitespace c = any (== c) horizontalWhitespaceChars

blankLine : PS (List Char)
blankLine = do
  cs <- line
  case all isHorizontalWhitespace cs of
    False => throw "nonblank line"
    True => pure cs

--*****************************************
--*             Inline syntax             *
--*****************************************

inlineElement : PS Inline

------------------------
-- Escaped Whitespace --
------------------------

hardLineBreak : PS Inline
hardLineBreak = do
  _ <- thisString "\\"
  _ <- spaces
  _ <- lineEnding
  pure $ HardLineBreak

nbsp : PS Inline
nbsp = do
  _ <- thisString "\\ "
  pure $ NonBreakingSpace

softLineBreak : PS Inline
softLineBreak = do
  _ <- lineEnding
  -- Check to see if the next line is empty, if it is, we are at the end of the inline
  -- content, go ahead and bail
  state <- save
  Nothing <- tryMaybe blankLine
    | Just _ => throw "End of inline"
  load state
  pure $ SoftLineBreak

---------------------
-- Emphasis/Strong --
---------------------

----------
-- Text --
----------

-- Process escape codes before processing as text
escapedText : PS Inline
escapedText = do
  c <- escapedChar
  pure $ Text c

-- Any non-line-ending character can be part of regular text
text : PS Inline
text = do
  c <- nonTerminal
  pure $ Text c

---------------------------
-- Overall Inline Parser --
---------------------------

inlineElement = oneOfE "" [
    hardLineBreak
  , nbsp
  , escapedText
  , softLineBreak
  -- Text is last so that anything can superseed it
  , text
  ]

inline : PS (List1 Inline)
inline = atLeastOne "Expected Inline Content" inlineElement


--*****************************************
--*           Utility Functions           *
--*****************************************

------------------
-- Constructors --
------------------

namespace Inline
  export
  fromString : String -> List (Inline)
  fromString str with (asList str)
    fromString "" | [] = []
    fromString (strCons c str) | (c :: x) =
      Text c :: fromString str | x

--*****************************************
--*              Unit Tests               *
--*****************************************

-------------------------------
-- Testing Utility Functions --
-------------------------------

-- Test a parser against a golden value with the supplied input
golden : Show a => Eq a =>
  (input : String) -> (reference : a) -> (parser : PS a) -> IO Bool
golden input ref parser = do
  putStrLn "Input: \{input}"
  putStrLn "Expected: \{show ref}"
  let got = runPS parser input
  case got of
    Left err => do
      putStrLn "Failed with error: \{err}"
      pure False
    Right x => do
      putStrLn "Output: \{show x}"
      pure $ x == ref

-------------------------
-- Inline Syntax Tests --
-------------------------

-- @@test Plain Text Hello World
inlineTextSmoke : IO Bool
inlineTextSmoke =
  let input = "Hello World!" in
  golden input (map Text . unpack $ input) (map forget inline)

-- @@test Escaped Text
inlineEscapedSmoke : IO Bool
inlineEscapedSmoke =
  let input = "Hello\\n\\*World"
      ref = fromString "Hello" ++ [Text '\n', Text '*'] ++ fromString "World"
  in golden input ref (map forget inline)

-- @@test Hard Line Break
inlineHardBreakSmoke : IO Bool
inlineHardBreakSmoke =
  let input = "Hello\\\nWorld"
      ref = fromString "Hello" ++ [HardLineBreak] ++ fromString "World"
  in golden input ref (map forget inline)

-- @@test Soft Line Break
inlineSoftBreakSmoke : IO Bool
inlineSoftBreakSmoke =
  let input = "Hello\nWorld"
      ref = fromString "Hello" ++ [SoftLineBreak] ++ fromString "World"
  in golden input ref (map forget inline)

-- @@test Nonbreaking Space
inlineNbspSmoke : IO Bool
inlineNbspSmoke =
  let input = "Hello\\ World"
      ref = fromString "Hello" ++ [NonBreakingSpace] ++ fromString "World"
  in golden input ref (map forget inline)
